
# This file was generated by PyScaffold from the git
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.

version_version = '0.1.3'
version_full = '668a064a916785631e724f74f9a509816410d31f'
def get_versions(default=dict(), verbose=False):
    return dict(version=version_version, full=version_full)
